/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.jbi.JBIException;
import javax.management.InstanceNotFoundException;
import javax.management.MBeanRegistrationException;
import javax.management.MBeanServer;
import javax.management.ObjectName;

import org.ow2.petals.component.framework.api.Constants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.util.PropertiesUtil;
import org.ow2.petals.component.framework.jbidescriptor.CDKJBIDescriptorException;
import org.ow2.petals.component.framework.jbidescriptor.JBIDescriptorBuilder;
import org.ow2.petals.component.framework.jbidescriptor.generated.SettableanyURI;
import org.ow2.petals.component.framework.mbean.MBeanHelper;
import org.ow2.petals.component.framework.util.StringHelper;

/**
 * A default bootstrap. This bootstrap class register an MBean object into the
 * JMX server that will be available only during the bootstrap phase. The
 * developer only needs to override methods {@code getMethodList()} to set the
 * list of method that he want to expose via JMX.
 * 
 * @author chamerling - EBM WebSourcing
 * @author rnaudin - EBM WebSourcing
 * 
 */
public class DefaultBootstrap extends JBIBootstrap {

    /**
     * The bootstrap MBean name
     */
    protected ObjectName mbeanName;

    /**
     * Returns the extension MBean name.
     * 
     * @return Returns the extension MBean name, except if the component is
     *         installed (ie. the extension MBean is not registered) in which
     *         case <code>null</code> is returned.
     * 
     * @see javax.jbi.component.ComponentLifeCycle#getExtensionMBeanName()
     */
    @Override
    public ObjectName getExtensionMBeanName() {

        final ObjectName objName;

        final MBeanServer server = this.installContext.getContext().getMBeanServer();
        if (server.isRegistered(this.mbeanName)) {
            objName = this.mbeanName;
        } else {
            objName = null;
        }

        return objName;
    }

    /**
     * Installs the component persisting the configuration done via the
     * extension Mbean.
     * 
     * @see javax.jbi.component.Bootstrap#onInstall()
     */
    @Override
    public void onInstall() throws JBIException {
        final File jbiXmlFile = new File(this.getInstallContext().getInstallRoot() + File.separator
                + Constants.Component.META_INF + File.separator + Constants.Component.JBI_XML);

        this.validateComponentSettableConfiguration();

        try {
            final FileOutputStream jbiOutputStream = new FileOutputStream(jbiXmlFile);
            try {
                JBIDescriptorBuilder.buildXmlJBIdescriptor(this.getJbiComponentConfiguration(),
                        jbiOutputStream);
            } finally {
                jbiOutputStream.close();
            }
        } catch (final IOException e) {
            throw new JBIException("Failed to persist the installation configuration : "
                    + e.getMessage());
        } catch (final CDKJBIDescriptorException e) {
            throw new JBIException("Failed to persist the installation configuration : "
                    + e.getMessage());
        }
    }

    /**
     * Clean resources allocated by the bootstrap
     * 
     * @see javax.jbi.component.Bootstrap#cleanUp()
     */
    @Override
    public void cleanUp() throws JBIException {

        MBeanServer server = installContext.getContext().getMBeanServer();
        if (server.isRegistered(mbeanName)) {
            try {
                server.unregisterMBean(mbeanName);
            } catch (InstanceNotFoundException e) {
                throw new JBIException("Error during bootstrap clean up : " + e.getMessage());
            } catch (MBeanRegistrationException e) {
                throw new JBIException("Error during bootstrap clean up" + e.getMessage());
            }
        }
    }

    /**
     * Initialize the bootstrap registering its Extension MBean into the JMX
     * server.
     * 
     * @see org.ow2.petals.component.framework.JBIBootstrap#doInit()
     */
    @Override
    protected void doInit() throws JBIException {

        this.mbeanName = this.installContext.getContext().getMBeanNames()
                .createCustomComponentMBeanName(
                        "bootstrap_" + this.installContext.getComponentName());

        MBeanHelper.registerMBean(this.installContext.getContext().getMBeanServer(),
                this.mbeanName, new ExtensionMBean(this));
    }

    /**
     * Method to be overridden if the component implementation needs to expose
     * methods via JMX. This method is called by the dynamic MBean.
     * 
     * @return a list of method names to expose
     * 
     */
    protected List<String> getMethodList() {
        return new ArrayList<String>();
    }

    /**
     * Method to be overridden if the component implementation needs to expose
     * attributes via JMX. This method is called by the dynamic MBean.
     * 
     * @return a list of attributes names to expose
     * 
     */
    protected List<String> getAttributeList() {
        return new ArrayList<String>();
    }

    /**
     * Validate the Component settable configuration.
     * 
     * @throws PEtALSCDKException
     */
    private void validateComponentSettableConfiguration() throws JBIException {
        SettableanyURI propertiesFile = this.getJbiComponentConfiguration().getComponent()
                .getPropertiesFile();

        if (propertiesFile != null && !StringHelper.isNullOrEmpty(propertiesFile.getValue())) {
            try {
                PropertiesUtil.loadProperties(propertiesFile.getValue());
            } catch (PEtALSCDKException e) {
                throw new JBIException(e.getMessage());
            }
        }
    }
}
