/**
 * PETALS - PETALS Services Platform. Copyright (c) 2006 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id: $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.process;

import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessageExchange;
import javax.management.AttributeChangeNotification;
import javax.management.Notification;
import javax.management.NotificationListener;

import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.mbean.RuntimeConfigurationNotifier;

/**
 * This class helps a Component to listen to {@link MessageExchange}s on the
 * DeliveryChannel (accept() method). It starts or stops a pool of thread that
 * listen to new incoming JBI from the bus. {@link MessageExchange
 * }s
 * <p/>
 * 
 * @author alouis - EBM Websourcing
 * @author ofabre - EBM Websourcing
 * @since 3.0
 */
public class JBIAcceptorManager {

    private Logger logger;

    private ThreadPoolExecutor acceptorPool;

    private JBIProcessorManager jbiProcessorManager;

    private AbstractComponent component;

    /**
     * The Acceptor factory used to create new Acceptors
     * 
     * @author Christophe HAMERLING - eBM WebSourcing
     * @since 3.0
     * 
     */
    private class JBIAcceptorThreadFactory implements ThreadFactory {

        private final AtomicInteger threadNumber = new AtomicInteger(1);

        private ThreadGroup group;

        public JBIAcceptorThreadFactory() {
            this.group = new ThreadGroup(Thread.currentThread().getThreadGroup(),
                    JBIAcceptorManager.this.component.getContext().getComponentName()
                            + "-JBIListener");
        }

        /*
         * (non-Javadoc)
         * 
         * @see java.util.concurrent.ThreadFactory#newThread(java.lang.Runnable)
         */
        public Thread newThread(Runnable r) {
            String name = JBIAcceptorManager.this.component.getContext().getComponentName()
                    + " -JBI Acceptor Thread #"
                    + Integer.toString(this.threadNumber.getAndIncrement());

            JBIAcceptorManager.this.logger.log(Level.FINE,
                    "Creating a new thread for JBIAcceptor Factory : " + name);

            MessageExchangeAcceptor meaProcessor = new MessageExchangeAcceptor(component,
                    JBIAcceptorManager.this.jbiProcessorManager.getThreadPool(),
                    JBIAcceptorManager.this.jbiProcessorManager.getObjectPool(),
                    JBIAcceptorManager.this.logger, this.group, name);
            meaProcessor.setDaemon(true);
            return meaProcessor;
        }
    }

    /**
     * Inner Notification Listener class.
     * 
     * @author rnaudin
     */
    private class RuntimeConfigurationListener implements NotificationListener {
        /*
         * (non-Javadoc)
         * 
         * @see
         * javax.management.NotificationListener#handleNotification(javax.management
         * .Notification, java.lang.Object)
         */
        public void handleNotification(Notification notification, Object obj) {
            if (notification instanceof AttributeChangeNotification) {
                AttributeChangeNotification attrNotif = (AttributeChangeNotification) notification;
                String attrName = attrNotif.getAttributeName();
                JBIAcceptorManager.this.logger.log(Level.FINE,
                        "Notification received for attribute : " + attrName);
                if (attrName.equals("acceptorPoolSize")) {
                    Integer newVal = (Integer) attrNotif.getNewValue();
                    JBIAcceptorManager.this.setCorePoolSize(newVal.intValue());
                }
            }
        }
    }

    /**
     * Creates a new instance of {@link JBIAcceptorManager}
     * 
     * @param channel
     * @param configuration
     *            The configuration of component
     * @param componentName
     *            Name of the component. Used to define the Thread name.
     * @param log
     * @param suManager
     * @param listener
     */
    public JBIAcceptorManager(AbstractComponent component,
            final JBIProcessorManager jbiProcessorManager,
            final RuntimeConfigurationNotifier runtimeConfiguration, Logger log) {
        super();
        this.component = component;
        this.jbiProcessorManager = jbiProcessorManager;
        this.logger = log;

        // set the listener on the Runtime Configuration Notifier
        runtimeConfiguration
                .addNotificationListener(new RuntimeConfigurationListener(), null, null);
    }

    /**
     * Initialize and start threads which listen to {@link MessageExchange}s
     * 
     * @throws JBIException
     * 
     */
    public void start() throws JBIException {
        // Note : since core-pool-size and maximum-pool-size are the same, the
        // pool size is fixed.
        this.acceptorPool = new ThreadPoolExecutor(this.component.getComponentConfiguration()
                .getAcceptorPoolSize().getValue(), this.component.getComponentConfiguration()
                .getAcceptorPoolSize().getValue(), 60, TimeUnit.SECONDS,
                new LinkedBlockingQueue<Runnable>(1024), new JBIAcceptorThreadFactory());
        this.acceptorPool.prestartAllCoreThreads();
    }

    /**
     * Stop threads that listen to {@link MessageExchange}s
     * 
     */
    public void stop() {
        // the executed task are hanging on the componentContext.accept()
        // method, so force the stop of these tasks directly by invoking
        // shutdownNow() instead of shutdown()
        this.acceptorPool.shutdownNow();
    }

    /**
     * Called on JMX notification.
     * 
     * @param threadCount
     */
    private void setCorePoolSize(int size) {
        if (this.acceptorPool != null) {
            this.logger.info("Reset the acceptor thread pool size to " + size);
            this.acceptorPool.setCorePoolSize(size);
        }
    }
}
