/**
 * PETALS - PETALS Services Platform. Copyright (c) 2008 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.process.async;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.pool.impl.GenericObjectPool;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.petals.component.framework.process.JBIProcessorManager;
import org.ow2.petals.component.framework.process.MessageExchangeProcessor;

/**
 * An asynchronous message Manager.
 * 
 * @author chamerling - EBM WebSourcing
 * @since 3.0
 * 
 */
public class AsyncMessageManager {

    public static final String ASYNC_MESSAGE_PROPERTY = "org.ow2.petals.component.framework.asynchronous";

    public static final String ASYNC_ORIGINAL_MESSAGE_PROPERTY = "org.ow2.petals.component.framework.asynchronous.original";

    public static final String ASYNC_EXPIRED_MESSAGE_PROPERTY = "org.ow2.petals.component.framework.asynchronous.expired";

    private final Logger logger;

    private Timer timer;

    private ThreadPoolExecutor jbiProcessorThreadPool;

    private GenericObjectPool jbiProcessorObjectPool;

    private final JBIProcessorManager jbiProcessorManager;

    private final Map<String, AsyncContext> asynchronousContexts;

    private final Map<String, AsyncContext> originalAsynchronousContexts;

    /**
     * 
     * @param logger
     */
    public AsyncMessageManager(final Logger logger, final JBIProcessorManager jbiProcessorManager) {
        this.logger = logger;
        this.asynchronousContexts = new ConcurrentHashMap<String, AsyncContext>();
        this.originalAsynchronousContexts = new ConcurrentHashMap<String, AsyncContext>();
        this.jbiProcessorManager = jbiProcessorManager;
    }

    /**
     * Start the Async Message Manager.
     */
    public void start() {
        //the time cannot be reused it must be recreated at each start
        this.timer = new Timer();
        this.timer.schedule(new AsyncContextCleaner(), 0, 2000);
        this.jbiProcessorThreadPool = this.jbiProcessorManager.getThreadPool();
        this.jbiProcessorObjectPool = this.jbiProcessorManager.getObjectPool();
    }

    /**
     * Stop the Async Message Manager.
     */
    public void stop() {
      //the time cannot be reused it must be destroyed at each stop
        this.timer.cancel();
        this.timer = null;
    }

    /**
     * Add an asynchronous message to the asynchronous messages map.
     * 
     * @param id
     * @param exchangeSynchronizer
     */
    public void addAsyncContext(final Exchange exchange, final AsyncContext asyncContext) {
        this.logger.fine("Add AsyncContext for exchange id: " + exchange.getExchangeId());
        exchange.setProperty(ASYNC_MESSAGE_PROPERTY, Boolean.toString(true));
        this.asynchronousContexts.put(exchange.getExchangeId(), asyncContext);
        if (!this.originalAsynchronousContexts.containsValue(asyncContext)) {
            asyncContext.getOriginalExchange().setProperty(ASYNC_ORIGINAL_MESSAGE_PROPERTY,
                    Boolean.toString(true));
            asyncContext.setStartTime(System.currentTimeMillis());
            this.originalAsynchronousContexts.put(asyncContext.getOriginalExchange()
                    .getExchangeId(), asyncContext);
        }
    }

    /**
     * Get an asynchronous Context by the exchange Id.
     * 
     * @param id
     * @return null if the asynchronous message is not found in the map
     */
    public AsyncContext getAsyncContext(final String id) {
        return this.asynchronousContexts.remove(id);
    }

    /**
     * Remove the asynchronous Context matching the given original Exchange Id.
     * 
     * @param id
     */
    public AsyncContext removeAsyncContext(final String id) {
        this.logger.fine("Remove Context with original exchange id : " + id);
        return this.originalAsynchronousContexts.remove(id);
    }

    /**
     * The cleaner task. Remove all the expired exchanges.
     */
    public class AsyncContextCleaner extends TimerTask {
        @Override
        public void run() {

            final List<AsyncContext> asyncContexts = new ArrayList<AsyncContext>();
            for (final Entry<String, AsyncContext> asyncContextEntry : AsyncMessageManager.this.originalAsynchronousContexts
                    .entrySet()) {
                if (asyncContextEntry.getValue().getExpirationTime() != 0
                        && System.currentTimeMillis() > asyncContextEntry.getValue()
                        .getExpirationTime()) {
                    asyncContextEntry.getValue().getOriginalExchange().setProperty(
                            ASYNC_EXPIRED_MESSAGE_PROPERTY, Boolean.toString(true));
                    asyncContexts.add(asyncContextEntry.getValue());
                }
            }

            final List<String> keys = new ArrayList<String>();
            for (final Entry<String, AsyncContext> asyncContextEntry : AsyncMessageManager.this.asynchronousContexts
                    .entrySet()) {
                if (asyncContextEntry.getValue().getExpirationTime() != 0
                        && System.currentTimeMillis() > asyncContextEntry.getValue()
                        .getExpirationTime()) {
                    keys.add(asyncContextEntry.getKey());
                }
            }
            for (final String id : keys) {
                AsyncMessageManager.this.asynchronousContexts.remove(id);
            }

            for (final AsyncContext asyncContext : asyncContexts) {
                try {
                    final MessageExchangeProcessor messageExchangeProcessor = (MessageExchangeProcessor) AsyncMessageManager.this.jbiProcessorObjectPool
                    .borrowObject();
                    messageExchangeProcessor.setMessageExchange((ExchangeImpl) asyncContext
                            .getOriginalExchange());
                    AsyncMessageManager.this.jbiProcessorThreadPool
                    .execute(messageExchangeProcessor);
                } catch (final Exception e) {
                    AsyncMessageManager.this.logger
                    .log(
                            Level.WARNING,
                            "Failed to borrow a Message Exchange Processor to process expired Asynchronous Context for original Exchange with id : "
                            + asyncContext.getOriginalExchange().getExchangeId(), e);
                }

            }
        }
    }
}
